/*
 * Copyright  2003 USC, Information Sciences Institute
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the
 * University of Southern California not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission. The University of Southern
 * California makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF
 * SOUTHERN CALIFORNIA BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl D. Worth <cworth@isi.edu>
 */

#include <cairo.h>

#include "write_png.h"

void
draw_outlines (cairo_t *cr, int surface_width, int surface_height);

void
draw_flat (cairo_t *cr, double width, double height);

void
draw_tent (cairo_t *cr, double width, double height);

void
draw_cylinder (cairo_t *cr, double width, double height);

cairo_pattern_t *
create_gradient (cairo_t *cr, double width, double height);

#define WIDTH 750
#define HEIGHT 500
#define STRIDE (WIDTH * 4)

char image[STRIDE*HEIGHT];

int
main (void)
{
    cairo_t *cr;

    cr = cairo_create ();

    cairo_set_target_image (cr, image, CAIRO_FORMAT_ARGB32,
		      WIDTH, HEIGHT, STRIDE);

    cairo_rectangle (cr, 0, 0, WIDTH, HEIGHT);
    cairo_set_rgb_color (cr, 1, 1, 1);
    cairo_fill (cr);

    draw_outlines (cr, WIDTH, HEIGHT);

    write_png_argb32 (image, "outline.png", WIDTH, HEIGHT, STRIDE);

    cairo_destroy (cr);

    return 0;
}

cairo_pattern_t *
create_gradient (cairo_t *cr, double width, double height)
{
    cairo_surface_t *gradient;
    cairo_matrix_t *matrix;
    cairo_pattern_t *gradient_pattern;

    cairo_save (cr);

    gradient = cairo_surface_create_similar (cairo_current_target_surface (cr),
					     CAIRO_FORMAT_ARGB32,
					     3, 2);
    cairo_set_target_surface (cr, gradient);

    cairo_set_rgb_color (cr, 0, 0, 0);
    cairo_rectangle (cr, 0, 0, 1, 2);
    cairo_fill (cr);

    cairo_set_rgb_color (cr, 1, 1, 1);
    cairo_rectangle (cr, 1, 0, 1, 2);
    cairo_fill (cr);

    cairo_set_rgb_color (cr, 0, 0, 0);
    cairo_rectangle (cr, 2, 0, 1, 2);
    cairo_fill (cr);

    cairo_restore (cr);

    matrix = cairo_matrix_create ();
    cairo_matrix_scale (matrix,
		   2.0 / width,
		   1.0 / height);
    
    gradient_pattern = cairo_pattern_create_for_surface (gradient);
    cairo_surface_destroy (gradient);
    cairo_pattern_set_matrix (gradient_pattern, matrix);
    cairo_pattern_set_filter (gradient_pattern, CAIRO_FILTER_BILINEAR);
    cairo_matrix_destroy (matrix);

    return gradient_pattern;
}

void
draw_outlines (cairo_t *cr, int surface_width, int surface_height)
{
    cairo_pattern_t *gradient;
    double width, height, pad;

    width = surface_width / 4.0;
    pad = (surface_width - (3 * width)) / 2.0;
    height = surface_height;

    gradient = create_gradient (cr, width, height);

    cairo_set_pattern (cr, gradient);
    draw_flat (cr, width, height);

    cairo_translate (cr, width + pad, 0);
    cairo_set_pattern (cr, gradient);
    draw_tent (cr, width, height);

    cairo_translate (cr, width + pad, 0);
    cairo_set_pattern (cr, gradient);
    draw_cylinder (cr, width, height);

    cairo_restore (cr);

    cairo_pattern_destroy (gradient);
}

void
draw_flat (cairo_t *cr, double width, double height)
{
    double hwidth = width / 2.0;

    cairo_rectangle (cr, 0, hwidth, width, height - hwidth);

    cairo_fill (cr);
}

void
draw_tent (cairo_t *cr, double width, double height)
{
    double hwidth = width / 2.0;

    cairo_move_to    (cr,       0,  hwidth);
    cairo_rel_line_to (cr,  hwidth, -hwidth);
    cairo_rel_line_to (cr,  hwidth,  hwidth);
    cairo_rel_line_to (cr,       0,  height - hwidth);
    cairo_rel_line_to (cr, -hwidth, -hwidth);
    cairo_rel_line_to (cr, -hwidth,  hwidth);
    cairo_close_path (cr);

    cairo_fill (cr);
}

void
draw_cylinder (cairo_t *cr, double width, double height)
{
    double hwidth = width / 2.0;

    cairo_move_to (cr, 0, hwidth);
    cairo_rel_curve_to (cr,
		      0, -hwidth,
		  width, -hwidth,
		  width, 0);
    cairo_rel_line_to (cr, 0, height - hwidth);
    cairo_rel_curve_to (cr,
		       0, -hwidth,
		  -width, -hwidth,
		  -width, 0);
    cairo_close_path (cr);

    cairo_fill (cr);
}
